#include "encoder_setting.h"
#include "rtc_base/refcountedobject.h"
#include "media/base/videocapturer.h"
#include "common_types.h"
#include "modules/video_coding/include/video_coding_defines.h"
#include "api/video_codecs/video_encoder.h"

namespace offcn
{
	const size_t EncoderSetting::kMaxNumberOfStreams;
	const int EncoderSetting::kMaxBitratePerStream[] = { 150 * 1000, 450 * 1000, 1500 * 1000 };
	const int EncoderSetting::kDefaultMinBitratePerStream[] = { 30 * 1000, 200 * 1000, 700 * 1000 };

	std::vector<webrtc::VideoStream> CreateVideoStream(int width, int height, const webrtc::VideoEncoderConfig& encoder_config)
	{
		std::vector<webrtc::VideoStream> stream_settings(encoder_config.number_of_streams);
		int bitrate_left_bps = encoder_config.max_bitrate_bps;

		for (size_t i = 0; i < encoder_config.number_of_streams; ++i)
		{
			stream_settings[i].width = (i + 1) * width / encoder_config.number_of_streams;
			stream_settings[i].height = (i + 1) * height / encoder_config.number_of_streams;
			stream_settings[i].max_framerate = 30;
			stream_settings[i].min_bitrate_bps = EncoderSetting::kDefaultMinBitratePerStream[i];

			int target_bitrate_bps = -1;
			int max_bitrate_bps = -1;
			// Use configured values instead of default values if values has been
			// configured.
			if (i < encoder_config.simulcast_layers.size())
			{
				const webrtc::VideoStream &stream = encoder_config.simulcast_layers[i];

				max_bitrate_bps = stream.max_bitrate_bps > 0 ? stream.max_bitrate_bps : EncoderSetting::kMaxBitratePerStream[i];
				max_bitrate_bps = std::min(bitrate_left_bps, max_bitrate_bps);

				target_bitrate_bps = stream.target_bitrate_bps > 0 ? stream.target_bitrate_bps : EncoderSetting::kMaxBitratePerStream[i];
				target_bitrate_bps = std::min(max_bitrate_bps, target_bitrate_bps);
			}
			else
			{
				max_bitrate_bps = std::min(bitrate_left_bps, EncoderSetting::kMaxBitratePerStream[i]);
				target_bitrate_bps = max_bitrate_bps;
			}

			RTC_DCHECK_NE(target_bitrate_bps, -1);
			RTC_DCHECK_NE(max_bitrate_bps, -1);
			stream_settings[i].target_bitrate_bps = target_bitrate_bps;
			stream_settings[i].max_bitrate_bps = max_bitrate_bps;
			stream_settings[i].max_qp = 56;

			if (i < encoder_config.simulcast_layers.size())
			{
				// Higher level controls are setting the active configuration for the
				// VideoStream.
				stream_settings[i].active = encoder_config.simulcast_layers[i].active;
			}
			else
			{
				stream_settings[i].active = true;
			}
			bitrate_left_bps -= stream_settings[i].target_bitrate_bps;
		}

		stream_settings[encoder_config.number_of_streams - 1].max_bitrate_bps += bitrate_left_bps;
		stream_settings[0].bitrate_priority = encoder_config.bitrate_priority;

		return stream_settings;
	}

	void FillEncoderConfiguration(webrtc::VideoCodecType codec_type, size_t num_streams, webrtc::VideoEncoderConfig* configuration)
	{
		configuration->codec_type = codec_type;
		configuration->number_of_streams = num_streams;
		configuration->video_stream_factory = new rtc::RefCountedObject<EncoderSetting>();
		configuration->max_bitrate_bps = 0;
		configuration->simulcast_layers = configuration->video_stream_factory.get()->CreateEncoderStreams(1280, 720, *configuration);
			//std::vector<webrtc::VideoStream>(configuratio);
		for (size_t i = 0; i < num_streams; ++i)
		{
			configuration->max_bitrate_bps += EncoderSetting::kMaxBitratePerStream[i];
		}
	}

	std::vector<webrtc::VideoStream>
		EncoderSetting::CreateEncoderStreams(int width, int height, const webrtc::VideoEncoderConfig &encoder_config)
	{
		return CreateVideoStream(width, height, encoder_config);
	}


	std::unique_ptr<cricket::VideoCapturer> OpenVideoCaptureDevice()
	{
		std::vector<std::string> device_names;
		{
			std::unique_ptr<webrtc::VideoCaptureModule::DeviceInfo> info(webrtc::VideoCaptureFactory::CreateDeviceInfo());
			if (!info)
			{
				return nullptr;
			}
			int num_devices = info->NumberOfDevices();
			for (int i = 0; i < num_devices; ++i)
			{
				const uint32_t kSize = 256;
				char name[kSize] = { 0 };
				char id[kSize] = { 0 };

				if (info->GetDeviceName(i, name, kSize, id, kSize) != -1)
				{
					device_names.push_back(name);
				}
			}
		}

		cricket::WebRtcVideoDeviceCapturerFactory factory;
		std::unique_ptr<cricket::VideoCapturer> capturer;
		for (const auto& name : device_names)
		{
			capturer = factory.Create(cricket::Device(name, 0));
			if (capturer)
			{
				break;
			}
		}
		return capturer;
	}

	void EncoderSetting::GetVideoCodecObj(webrtc::VideoCodec &codec, webrtc::VideoCodecType type)
	{
		memset(&codec, 0, sizeof(codec));

		codec.mode = webrtc::VideoCodecMode::kRealtimeVideo;
		codec.plType = 0;
		codec.numberOfSimulcastStreams = 1;

		codec.active = true;
		codec.timing_frame_thresholds = { webrtc::kDefaultTimingFramesDelayMs, webrtc::kDefaultOutlierFrameSizePercent };
		
		codec.width = 1280;
		codec.height = 720;
		codec.minBitrate = 30 * 1000 / 1000;
		codec.maxBitrate = 500 * 1000 / 1000;
		codec.qpMax = 56;

		webrtc::SimulcastStream *sim_stream = &codec.simulcastStream[0];
		sim_stream->width = 1280;
		sim_stream->height = 720;
		sim_stream->maxFramerate = 60;
		sim_stream->minBitrate = codec.minBitrate;
		sim_stream->maxBitrate = codec.maxBitrate;
		sim_stream->targetBitrate = 300 * 1000 / 1000;
		sim_stream->qpMax = codec.qpMax;
		sim_stream->numberOfTemporalLayers = 1;
		sim_stream->active = true;

		codec.codecType = webrtc::kVideoCodecH264;
		*codec.H264() = webrtc::VideoEncoder::GetDefaultH264Settings();
	}
};