#include "offcn_call.h"

#include "media/engine/adm_helpers.h"
#include "media/engine/apm_helpers.h"

#include "modules/audio_processing/audio_processing_impl.h"
#include "modules/audio_mixer/audio_mixer_impl.h"

#include "api/audio_codecs/audio_encoder_factory_template.h"
#include "api/audio_codecs/builtin_audio_encoder_factory.h"
#include "api/audio_codecs/builtin_audio_decoder_factory.h"
#include "api/video_codecs/builtin_video_decoder_factory.h"
#include "encoder_setting.h"
#include <iostream>

static void FillEncoderConfiguration(webrtc::VideoCodecType codec_type, size_t num_streams, webrtc::VideoEncoderConfig *configuration)
{
	configuration->codec_type = codec_type;
	configuration->number_of_streams = num_streams;
	configuration->video_stream_factory = new rtc::RefCountedObject<offcn::EncoderSetting>();
	configuration->max_bitrate_bps = 0;
	configuration->simulcast_layers = std::vector<webrtc::VideoStream>(num_streams);
	for (size_t i = 0; i < num_streams; ++i)
	{
		configuration->max_bitrate_bps += offcn::EncoderSetting::kMaxBitratePerStream[i];
	}
}

int OffcnCall::GetMicroPhoneIndexByName(rtc::scoped_refptr<webrtc::AudioDeviceModule> audio_device, const char *name)
{
    int num = audio_device->RecordingDevices();
    if(num <= 0)
    {
        return -1;
    }

    if(name == NULL || strlen(name) == 0)
    {
        return 0;
    }

    for (int i = 0; i < num; i++)
	{
		char tmp_name[128] = { 0 };
		char tmp_id[128] = { 0 };

		if (audio_device->RecordingDeviceName(i, tmp_name, tmp_id) < 0)
		{
			continue;
		}

		if (strcmp(tmp_name, name) == 0)
		{
            return i;
		}
	}
    return -1;
}
int OffcnCall::GetPlayoutIndexByName(rtc::scoped_refptr<webrtc::AudioDeviceModule> audio_device, const char *name)
{
    int num = audio_device->PlayoutDevices();
	if (num <= 0)
	{
		return -1;
	}

    if(name == NULL || strlen(name) == 0)
    {
        return 0;
    }

	for (int i = 0; i < num; i++)
	{
		char tmp_name[128] = { 0 };
		char tmp_id[128] = { 0 };

		if (audio_device->PlayoutDeviceName(i, tmp_name, tmp_id) < 0)
		{
			continue;
		}

		if (strcmp(tmp_name, name) == 0)
		{
            return i;
		}
	}
    return -1;
}

OffcnCall::OffcnCall()
{}
OffcnCall::~OffcnCall()
{
    if(call_)
    {
        call_.reset();
    }
}

bool OffcnCall::CreateCall(const char *playout_dev_name, const char *recorder_dev_name)
{
    //audio device module
    rtc::scoped_refptr<webrtc::AudioDeviceModule> audio_device_module = CreateAudioModule(playout_dev_name, recorder_dev_name);
    if(audio_device_module == nullptr)
    {
        return false;
    }

    //audio process
    webrtc::Config apm_config;
    webrtc::AudioProcessing *audio_processing = new rtc::RefCountedObject<webrtc::AudioProcessingImpl>(apm_config, nullptr, nullptr, nullptr, nullptr);
    if(audio_processing == nullptr)
    {
        return false;
    }
    if(audio_processing->Initialize() != webrtc::AudioProcessing::kNoError)
    {
        delete audio_processing;
        audio_processing = nullptr;
        return false;
    }

    //audio mixer
    rtc::scoped_refptr<webrtc::AudioMixer> audio_mixer = webrtc::AudioMixerImpl::Create();
    if(audio_mixer == nullptr)
    {
        return false;
    }

    //audio state
    webrtc::AudioState::Config  audio_state_config;
    audio_state_config.audio_device_module  = audio_device_module;
    audio_state_config.audio_processing     = audio_processing;
    audio_state_config.audio_mixer          = audio_mixer;
    rtc::scoped_refptr<webrtc::AudioState> audio_state = webrtc::AudioState::Create(audio_state_config);
    if(audio_state == nullptr)
    {
        return false;
    }
    
    //call
    webrtc::Call::Config call_config(&event_log_);
    call_config.audio_state = audio_state;

    audio_device_module->RegisterAudioCallback(audio_state->audio_transport());

    call_ = absl::WrapUnique(webrtc::Call::Create(call_config));
    if(call_ == nullptr)
    {
        return false;
    }

    return true;
}

rtc::scoped_refptr<webrtc::AudioDeviceModule> OffcnCall::CreateAudioModule(const char *playout_dev_name, const char *recorder_dev_name)
{
    rtc::scoped_refptr<webrtc::AudioDeviceModule> audio_module = webrtc::AudioDeviceModule::Create(webrtc::AudioDeviceModule::kPlatformDefaultAudio);
    if(audio_module == nullptr)
    {
        return nullptr;
    }

    webrtc::adm_helpers::Init(audio_module);

    int microphone_index = OffcnCall::GetMicroPhoneIndexByName(audio_module, recorder_dev_name);
    int playout_index = OffcnCall::GetPlayoutIndexByName(audio_module, playout_dev_name);
    if(microphone_index < 0 || playout_index < 0)
    {
        return nullptr;
    }
    
    int nRet = audio_module->SetRecordingDevice(microphone_index);
    if(nRet != 0)
    {
        return nullptr;
    }
    nRet = audio_module->SetPlayoutDevice(playout_index);
    if(nRet != 0)
    {
        return nullptr;
    }

    return audio_module;
}

webrtc::AudioSendStream *OffcnCall::CreateAudioSendStream(webrtc::Transport *transport, const char *cname)
{
    webrtc::AudioSendStream::Config config(transport);
    config.rtp.c_name = cname;
    config.rtp.nack.rtp_history_ms = 500;
    config.rtp.ssrc = 110;
	config.send_codec_spec = webrtc::AudioSendStream::Config::SendCodecSpec(110, { "opus", 48000, 2, {{"stereo", "1"}} });
    config.encoder_factory = webrtc::CreateBuiltinAudioEncoderFactory();
    
    webrtc::AudioSendStream *audio_stream = call_->CreateAudioSendStream(config);
    if(audio_stream == nullptr)
    {
        return nullptr;
    }
    
    return audio_stream;
	return nullptr;
}
webrtc::VideoSendStream *OffcnCall::CreateVideoSendStream(webrtc::VideoEncoderFactory *video_encoder_factory, webrtc::Transport *transport, const char *cname)
{
    webrtc::VideoSendStream::Config config(transport);
    config.rtp.ssrcs.push_back(112);
    config.rtp.nack.rtp_history_ms = 500;
    config.rtp.payload_name = "VP8";
    config.rtp.payload_type = 112;
    config.rtp.c_name = cname;
    config.encoder_settings.encoder_factory = video_encoder_factory;

    webrtc::VideoEncoderConfig encoder_config;
    FillEncoderConfiguration(webrtc::kVideoCodecVP8, 1, &encoder_config);
    encoder_config.max_bitrate_bps = 100 * 1000;
    encoder_config.video_format.name = "VP8";

    webrtc::VideoSendStream *video_stream = call_->CreateVideoSendStream(config.Copy(), encoder_config.Copy());
    if (video_stream == nullptr)
    {
        return nullptr;
    }

    return video_stream;
}

webrtc::AudioReceiveStream *OffcnCall::CreateAudioReceiveStream(webrtc::Transport *transport)
{
    webrtc::AudioReceiveStream::Config config;
    config.rtp.remote_ssrc = 110;
    config.rtp.local_ssrc  = 114;
    config.decoder_factory = webrtc::CreateBuiltinAudioDecoderFactory();
    config.decoder_map     = {{110, {"opus", 48000, 2}}};
    config.rtcp_send_transport = transport;
    return call_->CreateAudioReceiveStream(config);
}
webrtc::VideoReceiveStream *OffcnCall::CreateVideoReceiveStream(rtc::VideoSinkInterface<webrtc::VideoFrame> *videosink, webrtc::Transport *transport)
{
	webrtc::SdpVideoFormat format("VP8");
	std::unique_ptr<webrtc::VideoDecoderFactory> video_decoder_factory = webrtc::CreateBuiltinVideoDecoderFactory();
	video_decoder_ = video_decoder_factory->CreateVideoDecoder(format);

	webrtc::VideoReceiveStream::Config config(transport);
	config.renderer = videosink;//new FakeRenderer();
	config.rtp.local_ssrc = 911;
	config.rtp.remote_ssrc = 112;

	webrtc::VideoReceiveStream::Decoder decoder;
	decoder.decoder = video_decoder_.get();
	decoder.payload_name = "VP8";
	decoder.payload_type = 112;
	config.decoders.push_back(decoder);

	return call_->CreateVideoReceiveStream(config.Copy());
}